module subprogs
use,intrinsic :: iso_fortran_env
implicit none
    contains
    subroutine solute_flush(is_carbendazim, is_saturated, operation_loadingRate, operation_conc, error_data, pars)
    logical, intent(in) :: is_carbendazim, is_saturated
    real(real64), intent(in) :: pars(2)
    integer, intent(in) :: operation_loadingRate, operation_conc
    integer :: ITERMX= 100, NSTEPS= 1d9, NLIM= 6, i, conc_index, &
              ITER, IFLAG, stepN, NNODE_A, NNODE_A_1, NNODE_B, NNODE_B_1, &
              NNODE1_A, NNODE1_B, NNODE_A_static, NNODE_B_static, &
              NNODE_water, conf_data_size, &
              record_index, pesticide_type, saturation_condition
    integer, allocatable :: IMAT_A(:), IMAT_B(:)
    real(real64), intent(inout) ::  error_data
    real(real64) :: K(2),ALFA(2),GAMMAS(2),TETAS(2),TETAR(2), &
              EPS, ZMIN=0d0, DZ, DTMIN= 1.0d-6, DT, DTMAX, DMUL=1.3d0, &
              DDIV= 0.1d0, HB1, HBN, &
              DT0, DZ2, VWI_A, VWI_B, CWIN_A, CWIN_B, CWOU_A, CWOU_B, RATE, DHMAX_A, DHMAX_B, TIME, &
              TINIZ = 0.0d0, TFINAL,DTPRINT, TPRINT, & !DTPRINT=1d4, TPRINT= 6d4, &
              QWIN, QWOU, VWP, RATIO, BALER, HS, UH_A, UH_B, CONF_A, CONF_B, flush_limit , swithching_time, &
              RHO, ALPHA_L, air_boundary_thickness, &
              KD, DL_W, Ks, DG, KH, param_list_solute(11), param_list_water(4) 
    real(real64), allocatable :: Uwa_A(:), CON_A(:), CONO_A(:), H_A(:), Z_A(:), HO_A(:), &
                                 Awa_A(:), Bwa_A(:), Cwa_A(:), Rwa_A(:), Aso_A(:), Bso_A(:), Cso_A(:), Rso_A(:), &
                                 KM_A(:), CONF_data(:)
    real(real64), allocatable :: Uwa_B(:), CON_B(:), CONO_B(:), H_B(:), Z_B(:), HO_B(:), &
                                 Awa_B(:), Bwa_B(:), Cwa_B(:), Rwa_B(:), Aso_B(:), Bso_B(:), Cso_B(:), Rso_B(:), &
                                 KM_B(:)
    character(6) time_mark
    character(15) pesticide_list(2), saturation_condition_list(2)
    character(4) loadingRate_list(4), operationConc_list_car(5), operationConc_list_chl(7) 


    pesticide_list(:) = ['Carbendazim    ', 'Chlorothalonil ']
    saturation_condition_list(:) = ['Saturated      ', 'Unsaturated    ']
    loadingRate_list(:) =['L100', 'L200', 'L400', 'L050']

     if (is_carbendazim) then
        pesticide_type = 1
        operationConc_list_car(:) =['C100', 'C73 ', 'C43 ', 'C190', 'C127']

        ! 1: 100 ug/L, 2: 73 ug/L, 3: 43 ug/L, 4: 190 ug/L, 5: 127 ug/L
        if (operation_conc == 1) then
            CONF_A = 1.0d-7
        else if (operation_conc == 2) then
            CONF_A = 7.3d-8
        else if (operation_conc == 3) then
            CONF_A = 4.3d-8
        else if (operation_conc == 4) then
            CONF_A = 1.9d-7
        else if (operation_conc == 5) then
            CONF_A = 1.27d-7
        end if 

    else 
        pesticide_type = 2
        ! operationConc_list_chl(:) =['C10 ', 'C100', 'C500', 'C05G']
        operationConc_list_chl(:) =['C76 ', 'C79 ', 'C72 ', 'C236', 'C217', 'C391', 'C05G']

        ! 1: 76 ug/L, 2: 79 ug/L, 3: 72 ug/L, 
        ! 4: 236 ug/L, 5: 217 ug/L, 6: 392 ug/L, 7: 0.05 g/L 
        if (operation_conc == 1) then
            CONF_A = 7.6d-8 ! 76μg/L = 0.076 mg/L = 0.076 g/m3  7.6^-8 g/cm3
        else if (operation_conc == 2) then
            CONF_A = 7.9d-8 
        else if (operation_conc == 3) then
            CONF_A = 7.2d-8 
        else if (operation_conc == 4) then
            CONF_A = 2.36d-7 
        else if (operation_conc == 5) then
            CONF_A = 2.17d-7 
        else if (operation_conc == 6) then
            CONF_A = 3.92d-7 
        else if (operation_conc == 7) then
            CONF_A = 5.0d-5 ! 50mg/L = 0.05 g/L = 50 g/m3  5*10^-5 g/cm3
        end if 
    end if

    if (is_saturated) then
        saturation_condition = 1
        DZ = 1.0d0
        EPS = 7d-4
    else
        saturation_condition = 2
        DZ = 2.0d0
        EPS = 1.5d-16
    end if

    print*, 'Simulation starting...'
    print*, ''
    print*, 'Operational conditions for this simulation are: '  
    print*, ''  
    print*, 'Target pesticide: ', trim(pesticide_list(pesticide_type))
    print*, 'Saturation: ', trim(saturation_condition_list(saturation_condition))
    print*, 'Loading rate:    '//loadingRate_list(operation_loadingRate)(2:)//' L/d'
    if (is_carbendazim) then
        print*, 'Influent concentration:  '//trim(operationConc_list_car(operation_conc)(2:))//' ug/L'
    else
        print*, 'Influent concentration:  '//trim(operationConc_list_chl(operation_conc)(2:))//' ug/L'
    end if
    print*, ''

    NNODE_water = ceiling (5/DZ)+1
    
    ! Stage A
    NNODE_A = int(70/DZ)+1
    NNODE_A_1 = int(40/DZ)+1
    NNODE_A_static = NNODE_A - NNODE_water

    ! Stage B
    NNODE_B = int(90/DZ)+1
    NNODE_B_1 = int(60/DZ)+1
    NNODE_B_static = NNODE_B - NNODE_water

    
    ! Stage A
    allocate(IMAT_A(1:NNODE_A), Uwa_A(1:NNODE_A), CON_A(1:NNODE_A), CONO_A(1:NNODE_A), H_A(1:NNODE_A), Z_A(1:NNODE_A), &
      HO_A(1:NNODE_A), &
      Awa_A(1:NNODE_A), Bwa_A(1:NNODE_A), Cwa_A(1:NNODE_A), Rwa_A(1:NNODE_A), KM_A(1:NNODE_A-1) , &
      Aso_A(1:NNODE_A), Bso_A(1:NNODE_A), Cso_A(1:NNODE_A), Rso_A(1:NNODE_A))

    ! Stage B
    allocate(IMAT_B(1:NNODE_B), Uwa_B(1:NNODE_B), CON_B(1:NNODE_B), CONO_B(1:NNODE_B), H_B(1:NNODE_B), Z_B(1:NNODE_B), &
      HO_B(1:NNODE_B), &
      Awa_B(1:NNODE_B), Bwa_B(1:NNODE_B), Cwa_B(1:NNODE_B), Rwa_B(1:NNODE_B), KM_B(1:NNODE_B-1) , &
      Aso_B(1:NNODE_B), Bso_B(1:NNODE_B), Cso_B(1:NNODE_B), Rso_B(1:NNODE_B))

    
    ! Stage A
    Uwa_A = 0.0d0; Awa_A = 0.0d0; Bwa_A = 0.0d0; Cwa_A = 0.0d0; Rwa_A = 0.0d0 ; &
    Aso_A = 0.0d0; Bso_A = 0.0d0; Cso_A = 0.0d0; Rso_A = 0.0d0 

    ! Stage B
    Uwa_B = 0.0d0; Awa_B = 0.0d0; Bwa_B = 0.0d0; Cwa_B = 0.0d0; Rwa_B = 0.0d0 ; &
    Aso_B = 0.0d0; Bso_B = 0.0d0; Cso_B = 0.0d0; Rso_B = 0.0d0 

  
    ! 1=top layer 2=middle and bottom layer
    K = [0.5155d0, 1.1875d0] 
    ALFA = [0.12246d0, 0.18d0]
    GAMMAS = [2.8d0, 3.3d0] 
    TETAS = [0.37d0, 0.43d0]
    TETAR = [0.075d0, 0.04d0]

    RHO = 1.5d0 ! Bulk density
    air_boundary_thickness = 0.5d0 ! 5 mm = 0.5 cm
        
    TFINAL = 604001.1d0 !604800 sec equals to 1 week 
    DT= 1d-5
    DTPRINT=1d2 ! print interval
    TPRINT= 0d0

    conf_data_size = int(TFINAL/real(DTPRINT))+1
    record_index = 1

    allocate(CONF_data(1:conf_data_size))
    CONF_data = 0.0d0 ! zero reset

    if (operation_loadingRate == 1) then
        flush_limit = 6 
    else if (operation_loadingRate == 2) then
        flush_limit = 11
    else if (operation_loadingRate == 3) then
        flush_limit = 20
    else if (operation_loadingRate == 4) then
        flush_limit = 4
    end if

    swithching_time = 14400 ! 4 (hr) * 60 min * 60 sec
    
    Z_A(1) = ZMIN
    Z_B(1) = ZMIN

    do i = 2, NNODE_A
     Z_A(i) = Z_A(i-1) + dz
    end do

    do i = 2, NNODE_B
     Z_B(i) = Z_B(i-1) + dz
    end do

    KD = pars(1)
    Ks = pars(2)
   
    if (is_carbendazim) then
        ! KD=1.7d0  ! Distibution coefficient
        ! KD=0.58d0  ! calibrated
        DL_W=2.43d-6 ! Molecular diffusion coefficient in water
        ! Ks=7.0d-7 ! First-order degradation rate constant
        ! if (is_saturated) then
        !     ! Ks = 7d-7
        !     if (operation_loadingRate == 3) then ! HLR = 400 L/d
        !         Ks = 9.0d-6 ! calibrated
        !     else if (operation_loadingRate == 2) then
        !         KS = 4.5d-6
        !     else 
        !         Ks = 7.0d-7
        !         KD = 0.1d0
        !     end if
        ! else
        !     Ks=7d-7 
        ! end if
        DG=0d0 ! Molecular diffusion coefficient in air
        KH=1.28d-6 ! Henry constant
    else
        ! if (operation_loadingRate == 3) then ! HLR = 400 L/d
        !     KD = 6.5d0    ! from the literature
        ! else if (operation_loadingRate == 4) then !(operation_conc == 7) then ! Influent conc = 0.5 g/L
        !     ! KD=1.26d0 ! Distribution coefficient (from the literature)
        !     KD = 2.6d0 
        ! else 
        !     KD = 4.0d0 ! average value 4.0; 

        ! end if

        DL_W=1.39d-5 ! Molecular diffusion coefficient in water
        ! Ks=1.0d-6 ! First-order degradation rate constant
        DG=4.98d-2 ! Molecular diffusion coefficient in air
        KH=8.06d-6 ! Henry constant
    end if
        
    if (is_saturated) then
        ALPHA_L = 4.5d0     ! Dispersivity values from Giraldi et al. (2009)
        ! ALPHA_L = 130d0     ! calibration

        HO_A(:NNODE_water-1) = -0.1d0
        HO_A(NNODE_water) = 0.0d0
        do i = NNODE_water+1, NNODE_A
            HO_A(i) = 0d0 + 65d0/NNODE_A_static * dble(i-NNODE_water)
        end do

        HO_B(:NNODE_water-1) = -0.1d0
        HO_B(NNODE_water) = 0.0d0
        do i = NNODE_water+1, NNODE_B
            HO_B(i) = 0d0 + 85d0/NNODE_B_static * dble(i-NNODE_water)
        end do
    else
        ! ALPHA_L = 14d0       ! Dispersivity values from Giraldi et al. (2009)
        ALPHA_L =14d0       ! calibration
        
        HO_A(:) = -65d0
        HO_A(NNODE_A) = 0d0

        HO_B(:) = -65d0
        HO_B(NNODE_B) = 0d0
    end if
    
    CON_A(:) =  0d0
    CON_B(:) =  0d0

    CONO_A(:) =  0d0
    CONO_B(:) =  0d0
    
    IMAT_A(1:NNODE_A_1) = 1
    IMAT_A(NNODE_A_1+1:NNODE_A) = 2

    IMAT_B(1:NNODE_B_1) = 1
    IMAT_B(NNODE_B_1+1:NNODE_B) = 2

    VWI_A = 0.0d0 ! Initial volume of water present in the system [m^3]
    CWIN_A = 0.0d0 ! Cumulative water that entered the system
    CWOU_A = 0.0d0 ! Cumulative water that left the system

    VWI_B = 0.0d0 ! Initial volume of water present in the system [m^3]
    CWIN_B = 0.0d0 ! Cumulative water that entered the system
    CWOU_B = 0.0d0 ! Cumulative water that left the system
    
    
    do i = 1,NNODE_A   
     H_A(i) = HO_A(i)
     VWI_A = VWI_A + FUN_T(HO_A(i), ALFA, GAMMAS, TETAS, TETAR,  IMAT_A(i))*DZ 
    end do

    do i = 1,NNODE_B   
     H_B(i) = HO_B(i)
     VWI_B = VWI_B + FUN_T(HO_B(i), ALFA, GAMMAS, TETAS, TETAR,  IMAT_B(i))*DZ 
    end do


    RATE = 0.3 ! Fixed rate at first node[cm/sec]
    
    HB1 = 0.0d0; HBN = 0.0d0
    
    TIME = TINIZ
    DT0 = DT
    DZ2 = DZ*DZ

    param_list_water = [DZ, DZ2, HBN, RATE]

    param_list_solute = [DL_W, ALPHA_L, RATE, KH, KD, &
                        Ks, RHO, DG, DZ, DZ2, air_boundary_thickness]    

    Main_StageA: do stepN = 1, NSTEPS
     ITER = 1
     IFLAG = 0
     ! print*, 'Steps', stepN
     

    65 if(TIME < flush_limit  .or. mod(int(TIME), int(swithching_time)) < int(flush_limit)) then 

        DTMAX= 0.1d0
        ! if (is_saturated .neqv. .true.) then
            CALL ASSFDrate(param_list_water, DT, HO_A, H_A, Awa_A, Bwa_A, Cwa_A, Rwa_A, NNODE_A, KM_A, IMAT_A, &
                 ALFA, GAMMAS, K, TETAS, TETAR, is_saturated)  
            CALL TRIDAG(Awa_A, Bwa_A, Cwa_A, Rwa_A, Uwa_A, NNODE_A)
        ! end if

     else 
        DTMAX= 1.0d0
        ! if (is_saturated .neqv. .true.) then
            !print*, 'No Flush...'
            CALL ASSFD(param_list_water, DT, HO_A, H_A, Awa_A, Bwa_A, Cwa_A, Rwa_A, NNODE_A, KM_A, IMAT_A, &
                 ALFA, GAMMAS, K, TETAS, TETAR, is_saturated)  
            CALL TRIDAG(Awa_A, Bwa_A, Cwa_A, Rwa_A, Uwa_A, NNODE_A)
        ! end if
     end if
    
     DHMAX_A=1.0d-30
    
     do i = 1, NNODE_A
      UH_A = Uwa_A(i)/H_A(i)
          
      if (abs(UH_A) > DHMAX_A) then
        DHMAX_A = abs(UH_A)
      end if 

     end do

     ! Convergence test　
     ! EPS="max absolute ratio of the discart allowed for two subseqent iterations"
     if (DHMAX_A > EPS .AND. ITER < ITERMX) then 
          do i = 1, NNODE_A
           H_A(i) = H_A(i) + Uwa_A(i) 
          end do
          
          ITER = ITER + 1
          go to 65
    
     else if (DHMAX_A > EPS .AND. ITER >= ITERMX .AND. DT >= DTMIN) then  
          ! print*, 'Lower the time step'
          DT = DT*DDIV
          IFLAG = IFLAG + 1

          do i = 1, NNODE_A
           H_A(i) = H_A(i) + Uwa_A(i)
           H_A(i) = HO_A(i) + (H_A(i) - HO_A(i))*DT/DT0
          end do
           
          ITER = 1
          go to 65
      
     else if (DHMAX_A > EPS .AND. ITER >= ITERMX .AND. DT <= DTMIN) then
          print*, 'Convergence not reached'
          exit Main_StageA
      
     else if (DHMAX_A <= EPS) then
          TIME = TIME + DT
          ! print*, 'TIME: ', TIME
        
          if (TIME >= TPRINT) then
        ! update data
           do i = 1, NNODE_A
            H_A(i) = H_A(i) + Uwa_A(i)
           end do
          end if
                  
          if(TIME < flush_limit  .or. mod(int(TIME), int(swithching_time)) < int(flush_limit)) then
           ! DTMAX= 0.1d0
           CALL SOLUTEFDrate(param_list_solute, DT, H_A, CON_A, CONO_A, ALFA, GAMMAS, K, is_saturated, TETAS, TETAR, &
                Aso_A, Bso_A, Cso_A, Rso_A, NNODE_A, CONF_A, IMAT_A)
     
           CAll TRIDAG(Aso_A, Bso_A, Cso_A, Rso_A, CON_A, NNODE_A) 

           do conc_index = 1, size(CON_A)
            if(CON_A(conc_index) < 0.0d0) then
                CON_A(conc_index) = 0.0d0
            end if 
           end do

          else
           ! DTMAX= 1.0d0
           CALL SOLUTEFD(param_list_solute, DT, H_A, CON_A, CONO_A, ALFA, GAMMAS, K, is_saturated, TETAS, TETAR, &
                Aso_A, Bso_A, Cso_A, Rso_A, NNODE_A, IMAT_A)
     
           CAll TRIDAG(Aso_A, Bso_A, Cso_A, Rso_A, CON_A, NNODE_A) 

           do conc_index = 1, size(CON_A)
            if(CON_A(conc_index) < 0.0d0) then
                CON_A(conc_index) = 0.0d0
            end if 
           end do

          end if

          ! Mass balance
          NNODE1_A = NNODE_A - 1
        
          QWIN = KM_A(1)*((H_A(1)-H_A(2))/DZ + 1.0d0)
          QWOU = KM_A(NNODE1_A)*((H_A(NNODE1_A)-H_A(NNODE_A))/DZ + 1.0d0)
        
          CWIN_A = CWIN_A + QWIN*DT
          CWOU_A = CWOU_A + QWOU*DT
        
          if (TIME >= TPRINT) then

            ! Printing mass balance and writing results in csv files every 1000 sec
            if (mod(int(TIME),1000) == 0) then
              
               VWP = 0.0d0
            
               do i = 1, NNODE_A
                VWP = VWP + FUN_T(H_A(i), ALFA, GAMMAS, TETAS, TETAR, IMAT_A(i))*DZ
               end do 
              
               RATIO = (VWP-VWI_A)/(CWIN_A-CWOU_A)
               BALER = 100.0d0*abs(1.0d0 - RATIO)

               ! if (is_saturated .neqv. .true.) then
                   ! print*, '(Time >= TPRINT)', ' Time: ', Time
                   ! print*, 'Stage A mass balance: VWP, VWI, CWIN, CWOU, BALER, StepLevel'
                   ! write ( *, '(g14.6, g14.6, g14.6, g14.6, g14.6, I8)') &
                    ! VWP, VWI_A, CWIN_A, CWOU_A, BALER, stepN
                   ! print*, ''
               ! end if

               write(time_mark, '(I6.6)') int(TIME) ! unit: sec
               time_mark = adjustl(time_mark)


               ! if (is_carbendazim) then
               !     open(unit=16, &
               !      file='./output-pesticide-stageA-error/result.'&
               !          //pesticide_list(pesticide_type)(1:3)//'.'&
               !          //saturation_condition_list(saturation_condition)(1:3)//'.'&
               !          //loadingRate_list(operation_loadingRate)(:)//'.'&
               !          //trim(operationConc_list_car(operation_conc)(:))//'.'&
               !          //time_mark//'.csv')
               !  else 
               !      open(unit=16, &
               !      file='./output-pesticide-stageA-errors/result.'&
               !          //pesticide_list(pesticide_type)(1:3)//'.'&
               !          //saturation_condition_list(saturation_condition)(1:3)//'.'&
               !          //loadingRate_list(operation_loadingRate)(:)//'.'&
               !          //trim(operationConc_list_chl(operation_conc)(:))//'.'&
               !          //time_mark//'.csv')

               !  end if

               ! write(16, *) 'z, H, CON (ug/L), theta'
            
               ! do i = 1, NNODE_A
               !   write(16, 100) Z_A(i), H_A(i), CON_A(i)*10d9, FUN_T(H_A(i), ALFA, GAMMAS, TETAS, TETAR, IMAT_A(i))
               ! end do

            end if
               
               CONF_data(record_index) = CON_A(NNODE_A)
               record_index = record_index + 1
               ! print*, 'record_index: ', record_index

            ! 100 format(F8.4, ',', F12.2, ',', F20.2, ',', F8.4)
            ! close(16)
            
               TPRINT = TPRINT + DTPRINT
          end if 
    
          if (TIME >= TFINAL) then
               ! print*, 'TIME >= TFINAL'
               exit Main_StageA
          end if
    
          DT0 = DT
        
          if(ITER <= NLIM .AND. IFLAG == 0) then
           DT = DT*DMUL
          end if
        
          if (DT > DTMAX) then
           DT = DTMAX
          end if
        
          if (TIME + DT > TPRINT) then
           DT = TIME + DT - TPRINT
          end if
        
          do i = 1, NNODE_A
           HS = H_A(i) + (H_A(i) - HO_A(i))*DT/DT0
           HO_A(i) = H_A(i)
           H_A(i) = HS

           CONO_A(i) = CON_A(i)
          end do
     
    end if 
    
    end do Main_StageA
    print*, 'Exit main roop for Stage A'
    print*, ''



    if(is_carbendazim .eqv. .false.) then
        print*, 'Error being calculated for chlorothalonil...'

        if (operation_conc == 1) then
            print*, 'Effluent conc (Stage A): ', CON_A(NNODE_A)*10d9
            error_data = abs(CON_A(NNODE_A)*10d9 - 0.08d0)
        else if (operation_conc == 2) then    
            print*, 'Effluent conc (Stage A): ', CON_A(NNODE_A)*10d9
            error_data = abs(CON_A(NNODE_A)*10d9 - 1.43d0)
        else if (operation_conc == 3) then   
            print*, 'Effluent conc (Stage A): ', CON_A(NNODE_A)*10d9
            error_data = abs(CON_A(NNODE_A)*10d9 - 7.73d0)
        else if (operation_conc == 4) then
            print*, 'Effluent conc (Stage A): ', CON_A(NNODE_A)*10d9
            error_data = abs(CON_A(NNODE_A)*10d9 - 0.08d0)
        else if (operation_conc == 5) then
            print*, 'Effluent conc (Stage A): ', CON_A(NNODE_A)*10d9
            error_data = abs(CON_A(NNODE_A)*10d9 - 7.9d0)
        else if (operation_conc == 6) then
            print*, 'Effluent conc (Stage A): ', CON_A(NNODE_A)*10d9
            error_data = abs(CON_A(NNODE_A)*10d9 - 196d0)
        else if (operation_conc == 7) then
            print*, 'Effluent conc (Stage A): ', CON_A(NNODE_A)*10d9
            error_data = abs(CON_A(NNODE_A)*10d9 - 352d0)
        end if

        print*, 'Error for this simulation (chlorothalonil): ', error_data

    end if 


    ! print*, 'record_index: ', record_index

    TIME = TINIZ
    DT= 1d-5
    DT0 = DT
    TPRINT= 0d0


    Main_StageB: do stepN = 1, NSTEPS
     ITER = 1
     IFLAG = 0
     ! print*, 'Steps', stepN
     
    70 if(TIME < flush_limit  .or. mod(int(TIME), int(swithching_time)) < int(flush_limit)) then
        DTMAX= 0.1d0 
        ! if (is_saturated .neqv. .true.) then
            CALL ASSFDrate(param_list_water, DT, HO_B, H_B, Awa_B, Bwa_B, Cwa_B, Rwa_B, NNODE_B, KM_B, IMAT_B, &
               ALFA, GAMMAS, K, TETAS, TETAR, is_saturated)  
            CALL TRIDAG(Awa_B, Bwa_B, Cwa_B, Rwa_B, Uwa_B, NNODE_B)
        ! end if

     else 
        DTMAX= 1.0d0
        ! if (is_saturated .neqv. .true.) then
            !print*, 'No Flush...'
            CALL ASSFD(param_list_water, DT, HO_B, H_B, Awa_B, Bwa_B, Cwa_B, Rwa_B, NNODE_B, KM_B, IMAT_B, &
                ALFA, GAMMAS, K, TETAS, TETAR, is_saturated)  
            CALL TRIDAG(Awa_B, Bwa_B, Cwa_B, Rwa_B, Uwa_B, NNODE_B)
        ! end if
     end if
    
     DHMAX_B=1.0d-30

     do i = 1, NNODE_B
        UH_B = Uwa_B(i)/H_B(i)

        if (abs(UH_B) > DHMAX_B) then
            DHMAX_B = abs(UH_B)
        end if 
     end do

     ! Convergence test　
     ! EPS="max absolute ratio of the discart allowed for two subseqent iterations"
     if (DHMAX_B > EPS .AND. ITER < ITERMX) then 
          do i = 1, NNODE_B
           H_B(i) = H_B(i) + Uwa_B(i) 
          end do
          
          ITER = ITER + 1
          go to 70
      
    
     else if (DHMAX_B > EPS .AND. ITER >= ITERMX .AND. DT >= DTMIN) then  
          ! print*, 'Lower the time step'
          DT = DT*DDIV
          IFLAG = IFLAG + 1

          do i = 1, NNODE_B
           H_B(i) = H_B(i) + Uwa_B(i)
           H_B(i) = HO_B(i) + (H_B(i) - HO_B(i))*DT/DT0
          end do
           
          ITER = 1
          go to 70
      
     else if (DHMAX_B > EPS .AND. ITER >= ITERMX .AND. DT <= DTMIN) then
          print*, 'Convergence not reached'
          exit Main_StageB
      
     else if (DHMAX_B <= EPS) then
          TIME = TIME + DT
          ! print*, 'TIME: ', TIME
        
          if (TIME >= TPRINT) then
        ! update data
           do i = 1, NNODE_B
            H_B(i) = H_B(i) + Uwa_B(i)
           end do
          end if

         CONF_B = 0d0
         ! print*, 'int(TIME/DTPRINT)-1: ', ceiling(TIME/DTPRINT)-1
         CONF_B = CONF_data(ceiling(TIME/DTPRINT))

          if(TIME < flush_limit  .or. mod(int(TIME), int(swithching_time)) < int(flush_limit)) then 
               ! DTMAX= 0.1d0
               CALL SOLUTEFDrate(param_list_solute, DT, H_B, CON_B, CONO_B, ALFA, GAMMAS, K, is_saturated, TETAS, TETAR, &
                                 Aso_B, Bso_B, Cso_B, Rso_B, NNODE_B, CONF_B, IMAT_B)
         
               CAll TRIDAG(Aso_B, Bso_B, Cso_B, Rso_B, CON_B, NNODE_B) 

               do conc_index = 1, size(CON_B)
                if(CON_B(conc_index) < 0.0d0) then
                    CON_B(conc_index) = 0.0d0
                end if 
               end do

          else
               ! DTMAX= 1.0d0
               CALL SOLUTEFD(param_list_solute, DT, H_B, CON_B, CONO_B, ALFA, GAMMAS, K, is_saturated, TETAS, TETAR, &
                                Aso_B, Bso_B, Cso_B, Rso_B, NNODE_B, IMAT_B)
         
               CAll TRIDAG(Aso_B, Bso_B, Cso_B, Rso_B, CON_B, NNODE_B) 

               do conc_index = 1, size(CON_B)
                if(CON_B(conc_index) < 0.0d0) then
                    CON_B(conc_index) = 0.0d0
                end if 
               end do

          end if

          ! Mass balance
          NNODE1_B = NNODE_B - 1
        
          QWIN = KM_B(1)*((H_B(1)-H_B(2))/DZ + 1.0d0)
          QWOU = KM_B(NNODE1_B)*((H_B(NNODE1_B)-H_B(NNODE_B))/DZ + 1.0d0)
        
          CWIN_B = CWIN_B + QWIN*DT
          CWOU_B = CWOU_B + QWOU*DT
        
          if (TIME >= TPRINT) then

            if (mod(int(TIME),1000) == 0) then
              
               VWP = 0.0d0
            
               do i = 1, NNODE_B
                VWP = VWP + FUN_T(H_B(i), ALFA, GAMMAS, TETAS, TETAR, IMAT_B(i))*DZ
               end do 
              
               RATIO = (VWP-VWI_B)/(CWIN_B-CWOU_B)
               BALER = 100.0d0*abs(1.0d0 - RATIO)

               ! if (is_saturated .neqv. .true.) then
                   ! print*, '(Time >= TPRINT)', ' Time: ', Time
                   ! print*, 'Stage B mass balance: VWP, VWI, CWIN, CWOU, BALER, StepLevel'
                   ! write ( *, '(g14.6, g14.6, g14.6, g14.6, g14.6, I8)') &
                    ! VWP, VWI_B, CWIN_B, CWOU_B, BALER, stepN    
                   ! print*, ''
               ! end if
            
               write(time_mark, '(I6.6)') int(TIME) ! unit: sec
               time_mark = adjustl(time_mark)

               ! if (is_carbendazim) then
               !     open(unit=18, &
               !      file='./output-pesticide-stageB-error/result.'&
               !          //pesticide_list(pesticide_type)(1:3)//'.'&
               !          //saturation_condition_list(saturation_condition)(1:3)//'.'&
               !          //loadingRate_list(operation_loadingRate)(:)//'.'&
               !          //trim(operationConc_list_car(operation_conc)(:))//'.'&
               !          //time_mark//'.csv')
               !  else
               !      open(unit=18, &
               !      file='./output-pesticide-stageB-error/result.'&
               !          //pesticide_list(pesticide_type)(1:3)//'.'&
               !          //saturation_condition_list(saturation_condition)(1:3)//'.'&
               !          //loadingRate_list(operation_loadingRate)(:)//'.'&
               !          //trim(operationConc_list_chl(operation_conc)(:))//'.'&
               !          //time_mark//'.csv')

               !  end if

               ! write(18, *) 'z, H, CON (ug/L), theta'
            
               ! do i = 1, NNODE_B
               !   write(18, 110) Z_B(i), H_B(i), CON_B(i)*10d9, FUN_T(H_B(i), ALFA, GAMMAS, TETAS, TETAR, IMAT_B(i))
               ! end do
            end if


            ! 110 format(F8.4, ',', F12.2, ',', F20.2, ',', F8.4)
            ! close(18)
            
               TPRINT = TPRINT + DTPRINT
          end if 
    
          if (TIME >= TFINAL) then
               ! print*, 'TIME >= TFINAL'
               exit Main_StageB
          end if
    
          DT0 = DT
        
          if(ITER <= NLIM .AND. IFLAG == 0) then
           DT = DT*DMUL
          end if
        
          if (DT > DTMAX) then
           DT = DTMAX
          end if
        
          if (TIME + DT > TPRINT) then
           DT = TIME + DT - TPRINT
          end if
        
          do i = 1, NNODE_B
           HS = H_B(i) + (H_B(i) - HO_B(i))*DT/DT0
           HO_B(i) = H_B(i)
           H_B(i) = HS

           CONO_B(i) = CON_B(i)
          end do
     
    end if 
    
    end do Main_StageB

    print*, 'Exit main roop for Stage B'
    print*, ''

! exp_data = np.array([22.3, 8.3, 63, 8.67, 22.7, 32, 109.5])
    if(is_carbendazim) then
        print*, 'Error being calculated for carbendazim...'

        if (operation_conc == 1) then
            ! 100 L/d
            if(operation_loadingRate == 1) then
                print*, 'Effluent conc (Stage B): ', CON_B(NNODE_B)*10d9
                error_data = abs(CON_B(NNODE_B)*10d9 - 8.67d0)
            
            ! 200 L/d
            else if (operation_loadingRate == 2) then
                print*, 'Effluent conc (Stage B): ', CON_B(NNODE_B)*10d9
                error_data = abs(CON_B(NNODE_B)*10d9 - 22.7d0)
            
            ! 400 L/d
            else if (operation_loadingRate == 3) then
                print*, 'Effluent conc (Stage B): ', CON_B(NNODE_B)*10d9
                error_data = abs(CON_B(NNODE_B)*10d9 - 32d0)
            end if

        else if (operation_conc == 2) then
            if(operation_loadingRate == 1) then
                print*, 'Effluent conc (Stage B): ', CON_B(NNODE_B)*10d9
                error_data = abs(CON_B(NNODE_B)*10d9 - 22.3d0)
            end if 

        else if (operation_conc == 3) then
            if(operation_loadingRate == 2) then
                print*, 'Effluent conc (Stage B): ', CON_B(NNODE_B)*10d9
                error_data = abs(CON_B(NNODE_B)*10d9 - 8.3d0)
            end if 

        else if (operation_conc == 4) then
            if(operation_loadingRate == 3) then
                print*, 'Effluent conc (Stage B): ', CON_B(NNODE_B)*10d9
                error_data = abs(CON_B(NNODE_B)*10d9 - 63d0)
            end if 

        else if (operation_conc == 5) then
            if(operation_loadingRate == 2) then
                print*, 'Effluent conc (Stage B): ', CON_B(NNODE_B)*10d9
                error_data = abs(CON_B(NNODE_B)*10d9 - 109.5d0)
            end if 

        end if

    ! else 
    !     error_data = error_data + CON_B(NNODE_B)*10d9

        print*, 'Error for this simulation (carbendazim): ', error_data
    end if 

    ! print*, 'Effluent pesticide concentration: ', CON_B(NNODE_B)*10d9



    print*, 'Simulation completed...'
    print*, ''
    
    end subroutine

  SUBROUTINE TRIDAG(A,B,C,R,Output,N)
    real(real64), intent(in) :: A(:), B(:), C(:), R(:)
    real(real64), intent(out) :: Output(:)
    integer, intent(in) :: N
    real(real64) GAM(2:N), BET
    
    integer j
    
    GAM = 0.0d0
    
    if (B(1) == 0) then
     stop 'B(1) == 0'
    end if 
    
    BET = B(1) ! B(1) = 1
    Output(1) = R(1)/BET ! R(1) = 0
    
    do j = 2,N
      GAM(j) = C(j-1)/BET
      BET = B(j) - A(j) * GAM(j)
    
      if (BET == 0) then
       stop 'BET == 0'
      end if
     
      Output(j) = (R(j) - A(j)*Output(j-1))/BET
    end do
    
    do j = N-1, 1, -1
      Output(j) = Output(j)-GAM(j+1)*Output(j+1)
    end do
    
    end SUBROUTINE TRIDAG

    SUBROUTINE ASSFD(param_list, DT, HO,H,Awa,Bwa,Cwa,Rwa,N,KM, IMAT, &
                            ALFA, GAMMAS, K, TETAS, TETAR, is_saturated)
    real(real64), intent(in):: DT, HO(:), H(:), param_list(:), ALFA(:), GAMMAS(:), K(:), &
                                    TETAS(:), TETAR(:)
    integer, intent(in) :: N, IMAT(:)
    logical, intent(in) :: is_saturated
    real(real64), intent(out) :: Awa(:), Bwa(:), Cwa(:), Rwa(:), KM(1:N-1)
    real(real64) FK(1:N), CC, DZ, DZ2, HBN, RATE
    integer N1, i


    N1 = N-1
    DZ = param_list(1)
    DZ2 = param_list(2)
    HBN = param_list(3)
    RATE = param_list(4)
    

    do i = 1, N
     FK(i) = FUN_K(H(i), ALFA, GAMMAS, K,  IMAT(i), is_saturated)
    end do
    
    do i = 1, N1
     KM(i) = 0.5d0*(FK(i)+FK(i+1))
    end do
    
    do i = 2, N1
     CC = FUN_C(H(i), ALFA, GAMMAS, TETAS, TETAR, IMAT(i))
     Awa(i) = - KM(i-1)/DZ2
     Bwa(i) = CC/DT + (KM(i-1)+KM(i))/DZ2
     Cwa(i) = -KM(i)/DZ2
     Rwa(i) = (KM(i-1)*(H(i-1)-H(i)) + KM(i)*(H(i+1)-H(i)))/DZ2 - (KM(i)-KM(i-1))/DZ &
             - CC*(H(i)-HO(i))/DT 
    end do

    if (is_saturated) then
        CC = FUN_C(H(N),ALFA, GAMMAS, TETAS, TETAR, IMAT(N))
        Rwa(N) = KM(N1)*(H(N1)-H(N))/DZ2 + KM(N1)/DZ - CC*(H(N)-HO(N))/DT
        Bwa(N) = CC/DT + KM(N1)/DZ2
        Awa(N) = -KM(N1)/DZ2

        CC = FUN_C(H(1),ALFA, GAMMAS, TETAS, TETAR, IMAT(1))
        Bwa(1) = CC/DT + KM(1)/DZ2
        Cwa(1) = -KM(1)/DZ2
        Rwa(1) = KM(1)*(H(2)-H(1))/DZ2 - KM(1)/DZ - CC*(H(1)-HO(1))/DT

    else    
         Rwa(N)=HBN
         Bwa(N)=1.0d0
         Awa(N)=0.0d0
         Rwa(N1) = Rwa(N1) - Cwa(N1)*HBN
         Cwa(N1) = 0.0d0
         
         CC = FUN_C(H(1),ALFA, GAMMAS, TETAS, TETAR, IMAT(1))
         Bwa(1) = CC/DT 
         Cwa(1) = 0.0d0
         Rwa(1) = - KM(1)/DZ - CC*(H(1)-HO(1))/DT 
    end if

    end subroutine ASSFD


    SUBROUTINE ASSFDrate(param_list, DT, HO, H, Awa, Bwa, Cwa, Rwa, N, KM, IMAT, &
                            ALFA, GAMMAS, K, TETAS, TETAR, is_saturated)
        real(real64), intent(in):: DT, HO(:), H(:), param_list(:), ALFA(:), GAMMAS(:), K(:), &
                                    TETAS(:), TETAR(:)
        integer, intent(in) :: N, IMAT(:)
        logical, intent(in) :: is_saturated
        real(real64), intent(out) :: Awa(:), Bwa(:), Cwa(:), Rwa(:), KM(1:N-1)
        real(real64) FK(1:N), CC, DZ, DZ2, HBN, RATE
        integer N1, i
        
        N1 = N-1
        DZ = param_list(1)
        DZ2 = param_list(2)
        HBN = param_list(3)
        RATE = param_list(4)

        do i = 1, N
         FK(i) = FUN_K(H(i), ALFA, GAMMAS, K, IMAT(i), is_saturated)
        end do
        
        do i = 1, N1
         KM(i) = 0.5d0*(FK(i)+FK(i+1))
        end do
        
        do i = 2, N1
         CC = FUN_C(H(i), ALFA, GAMMAS, TETAS, TETAR, IMAT(i))
         Awa(i) = - KM(i-1)/DZ2
         Bwa(i) = CC/DT + (KM(i-1)+KM(i))/DZ2
         Cwa(i) = -KM(i)/DZ2
         Rwa(i) = (KM(i-1)*(H(i-1)-H(i)) + KM(i)*(H(i+1)-H(i)))/DZ2 - (KM(i)-KM(i-1))/DZ &
                - CC*(H(i)-HO(i))/DT 
        end do

        if (is_saturated) then
            CC = FUN_C(H(N),ALFA, GAMMAS, TETAS, TETAR, IMAT(N))
            Rwa(N) = KM(N1)*(H(N1)-H(N))/DZ2 + KM(N1)/DZ - CC*(H(N)-HO(N))/DT - RATE/DZ
            Bwa(N) = CC/DT + KM(N1)/DZ2
            Awa(N) = -KM(N1)/DZ2
        else
             Rwa(N)=HBN
             Bwa(N)=1.0d0
             Awa(N)=0.0d0
             Rwa(N1) = Rwa(N1) - Cwa(N1)*HBN
             Cwa(N1) = 0.0d0
        end if

         CC = FUN_C(H(1), ALFA, GAMMAS, TETAS, TETAR, IMAT(1))
         Bwa(1) = CC/DT + KM(1)/DZ2
         Cwa(1) = -KM(1)/DZ2
         Rwa(1) = KM(1)*(H(2)-H(1))/DZ2 - KM(1)/DZ - CC*(H(1)-HO(1))/DT + RATE/DZ
        
    end subroutine ASSFDrate


    SUBROUTINE SOLUTEFD(param_list, DT, H,CON, CONO, ALFA, GAMMAS, K, is_saturated, TETAS, TETAR, &
                        Aso,Bso,Cso,Rso,N, IMAT)
        real(real64), intent(in):: H(:),CON(:), CONO(:), param_list(:), DT
        real(real64), intent(in) :: ALFA(:), GAMMAS(:), K(:), TETAS(:), TETAR(:)
        integer, intent(in) :: N, IMAT(:)
        logical, intent(in) :: is_saturated
        real(real64), intent(out) :: Aso(:), Bso(:), Cso(:), Rso(:)
        real(real64) FT(1:N), FK(1:N), S(N),R(N),air(N), ELV(N), ESL(N), EQ(N), DL(N)
        real(real64) KM(1:N-1), vertical_flux(N)
        real(real64) DL_W, ALPHA_L, RATE, KH, KD, Ks, RHO, DG, DZ, DZ2, air_boundary_thickness
        integer N1, i
    
        N1 = N-1

        DL_W = param_list(1)
        ALPHA_L = param_list(2)
        RATE = param_list(3)
        KH = param_list(4)
        KD = param_list(5)
        Ks = param_list(6)
        RHO = param_list(7)
        DG = param_list(8)
        DZ = param_list(9)
        DZ2 = param_list(10)
        air_boundary_thickness = param_list(11)

        if(is_saturated)then

            ! do i = 1, NNODE_water
            !     FK(i) = FUN_K(H(i), ALFA, GAMMAS, K, IMAT(i), is_saturated)
            !     FT(i) = FUN_T(H(i), ALFA, GAMMAS, TETAS, TETAR, IMAT(i))
            ! end do

            ! do i = 1, NNODE_water
            !     KM(i) = 0.5d0*(FK(i)+FK(i+1))
            ! end do

            ! do i = 2, NNODE_water
            !     vertical_flux(i) =  (-KM(i)*((H(i+1)-H(i))/DZ - 1d0) -KM(i-1)*((H(i)-H(i-1))/DZ - 1d0))/2.0d0
            ! end do
            ! vertical_flux(1) = -KM(1)*((H(2)-H(1))/DZ - 1d0) 
            ! vertical_flux(NNODE_water) = -KM(NNODE_water-1)*((H(N)-H(NNODE_water-1))/DZ - 1d0) 

            do i = 1, N
              FT(i) = FUN_T(H(i), ALFA, GAMMAS, TETAS, TETAR, IMAT(i))
              DL(i) = DL_W * (FT(i)**(7d0/3d0))/TETAS(IMAT(i))**2.0d0
              S(i)=KD*CON(i)

              air(i)= TETAS(IMAT(i)) - FT(i)

              R(i) = FT(i) + air(i)*KH + RHO*KD

              ELV(i) = air(i)*DG*KH/DZ2
              ESL(i) = FT(i)*DL(i)/DZ2
              EQ(i) = 0.0d0
            end do

            do i = 2, N1
             Aso(i) = -ELV(i)*DT/R(i) - ESL(i)*DT/R(i) - EQ(i)*DT
             Bso(i) = 2d0*ELV(i)*DT/R(i) + 2d0*ESL(i)*DT/R(i) + Ks*DT + 1d0
             Cso(i) = -ELV(i)*DT/R(i) - ESL(i)*DT/R(i) + EQ(i)*DT
             Rso(i) = CONO(i)
            end do      

             Aso(N) = -1d0
             Bso(N) = 1d0
             Rso(N) = 0d0

             Bso(1) = (air(1)*DG*KH + FT(1)*DL(1))/DZ  + air(1)*DG*KH/air_boundary_thickness 
             Cso(1) = - (air(1)*DG*KH + FT(1)*DL(1))/DZ 
             Rso(1) = 0d0

        else
            do i = 1, N
                FK(i) = FUN_K(H(i), ALFA, GAMMAS, K, IMAT(i), is_saturated)
                FT(i) = FUN_T(H(i), ALFA, GAMMAS, TETAS, TETAR, IMAT(i))
            end do

            do i = 1, N1
                KM(i) = 0.5d0*(FK(i)+FK(i+1))
            end do

            do i = 2, N1
                vertical_flux(i) =  (-KM(i)*((H(i+1)-H(i))/DZ - 1d0) -KM(i-1)*((H(i)-H(i-1))/DZ - 1d0))/2.0d0
            end do
            vertical_flux(1) = -KM(1)*((H(2)-H(1))/DZ - 1d0) 
            vertical_flux(N) = -KM(N1)*((H(N)-H(N1))/DZ - 1d0) 

            do i = 1, N
              DL(i) = DL_W * (FT(i)**(7d0/3d0))/TETAS(IMAT(i))**2.0d0+ ALPHA_L * vertical_flux(i)/FT(i)
              S(i)=KD*CON(i) 

              air(i)= TETAS(IMAT(i)) - FT(i)

              R(i) = FT(i) + air(i)*KH + RHO*KD

              ELV(i) = air(i)*DG*KH/DZ2
              ESL(i) = FT(i)*DL(i)/DZ2

              EQ(i) = vertical_flux(i)/(2d0*DZ) 
            end do
            
            do i = 2, N1         
             Aso(i) = -ELV(i)*DT/R(i) - ESL(i)*DT/R(i) - EQ(i)*DT/R(i)
             Bso(i) = 2d0*ELV(i)*DT/R(i) + 2d0*ESL(i)*DT/R(i) + Ks*DT + 1
             Cso(i) = -ELV(i)*DT/R(i) - ESL(i)*DT/R(i) + EQ(i)*DT/R(i)
             Rso(i) = CONO(i)
            end do        
             Aso(N) = -1d0
             Bso(N) = 1d0
             Rso(N) = 0d0

             Bso(1) = (air(1)*DG*KH + FT(1)*DL(1))/DZ  + DG*KH/air_boundary_thickness  + vertical_flux(1) 
             Cso(1) = -(air(1)*DG*KH + FT(1)*DL(1))/DZ 
             Rso(1) = 0d0
        end if

        end subroutine SOLUTEFD

        SUBROUTINE SOLUTEFDrate(param_list, DT, H,CON, CONO, ALFA, GAMMAS, K, is_saturated, TETAS, TETAR, &
                                Aso,Bso,Cso,Rso,N, CONF, IMAT)
        real(real64), intent(in):: H(:),CON(:), CONO(:), CONF, param_list(:), DT
        real(real64), intent(in) :: ALFA(:), GAMMAS(:), K(:), TETAS(:), TETAR(:)
        integer, intent(in) :: N, IMAT(:)
        logical, intent(in) :: is_saturated
        real(real64), intent(out) :: Aso(:), Bso(:), Cso(:), Rso(:)
        real(real64) FT(1:N),FK(1:N), S(N),R(N),air(N), ELV(N), ESL(N), EQ(N), DL(N)
        real(real64) KM(1:N-1), vertical_flux(N)
        real(real64) DL_W, ALPHA_L, RATE, KH, KD, Ks, RHO, DG, DZ, DZ2, air_boundary_thickness
        integer N1, i
    
        N1 = N-1
  
        DL_W = param_list(1)
        ALPHA_L = param_list(2)
        RATE = param_list(3)
        KH = param_list(4)
        KD = param_list(5)
        Ks = param_list(6)
        RHO = param_list(7)
        DG = param_list(8)
        DZ = param_list(9)
        DZ2 = param_list(10)
        air_boundary_thickness = param_list(11)


        if (is_saturated) then
            do i = 1, N
              FT(i) = FUN_T(H(i), ALFA, GAMMAS, TETAS, TETAR, IMAT(i))

              DL(i) = DL_W * (FT(i)**(7d0/3d0))/TETAS(IMAT(i))**2.0d0 + ALPHA_L * (RATE/FT(i))
              S(i)=KD*CON(i) 

              air(i)= TETAS(IMAT(i)) - FT(i)

              R(i) = FT(i) + air(i)*KH + RHO*KD

              ELV(i) = air(i)*DG*KH/DZ2
              ESL(i) = FT(i)*DL(i)/DZ2

              EQ(i) = RATE/(2d0*DZ) 
            end do

            do i = 2, N1
             Aso(i) = -ELV(i)*DT/R(i) - ESL(i)*DT/R(i) - EQ(i)*DT
             Bso(i) = 2d0*ELV(i)*DT/R(i) + 2d0*ESL(i)*DT/R(i) + Ks*DT + 1
             Cso(i) = -ELV(i)*DT/R(i) - ESL(i)*DT/R(i) + EQ(i)*DT
             Rso(i) = CONO(i)
            end do        
             Aso(N) = -1d0
             Bso(N) = 1d0
             Rso(N) = 0d0


             Bso(1) = (air(1)*DG*KH + FT(1)*DL(1))/DZ  + DG*KH/air_boundary_thickness  + RATE            
             Cso(1) = - (air(1)*DG*KH + FT(1)*DL(1))/DZ 
             Rso(1) = RATE * CONF

        else 
            do i = 1, N
                FK(i) = FUN_K(H(i), ALFA, GAMMAS, K, IMAT(i), is_saturated)
                FT(i) = FUN_T(H(i), ALFA, GAMMAS, TETAS, TETAR, IMAT(i))
            end do

            do i = 1, N1
                KM(i) = 0.5d0*(FK(i)+FK(i+1))
            end do


            do i = 2, N1
                vertical_flux(i) =  (-KM(i)*((H(i+1)-H(i))/DZ - 1d0) + -KM(i-1)*((H(i)-H(i-1))/DZ - 1d0))/2.0d0
            end do
            vertical_flux(1) = -KM(1)*((H(2)-H(1))/DZ - 1d0) + RATE
            vertical_flux(N) = -KM(N1)*((H(N)-H(N1))/DZ - 1d0) 
        
            do i = 1, N
              DL(i) = DL_W * (FT(i)**(7d0/3d0))/TETAS(IMAT(i))**2.0d0 + ALPHA_L * (vertical_flux(i)/FT(i))
              S(i)=KD*CON(i) 

              air(i)= TETAS(IMAT(i)) - FT(i)

              R(i) = FT(i) + air(i)*KH + RHO*KD

              ELV(i) = air(i)*DG*KH/DZ2
              ESL(i) = FT(i)*DL(i)/DZ2
              EQ(i) = vertical_flux(i)/(2d0*DZ) 

            end do

            do i = 2, N1
             Aso(i) = -ELV(i)*DT/R(i) - ESL(i)*DT/R(i) - EQ(i)*DT/R(i)
             Bso(i) = 2d0*ELV(i)*DT/R(i) + 2d0*ESL(i)*DT/R(i) + Ks*DT + 1d0
             Cso(i) = -ELV(i)*DT/R(i) - ESL(i)*DT/R(i) + EQ(i)*DT/R(i)
             Rso(i) = CONO(i)
            end do        
             Aso(N) = -1d0
             Bso(N) = 1d0
             Rso(N) = 0d0

             Bso(1) = (air(1)*DG*KH + FT(1)*DL(1))/DZ  + air(1)*DG*KH/air_boundary_thickness  + vertical_flux(1)  
             Cso(1) = - (air(1)*DG*KH + FT(1)*DL(1))/DZ 
             Rso(1) = vertical_flux(1) * CONF
        end if
        end subroutine SOLUTEFDrate
    
    
    
    function FUN_K(X, ALFA, GAMMAS, K, IM, is_saturated) result(conductivity)
    real(real64), intent(in) :: X, ALFA(:), GAMMAS(:), K(:)
    integer, intent(in) :: IM 
    logical, intent(in) :: is_saturated
    real(real64) conductivity
    real(real64) N, M, B, ALFN, SE 
    ! real(real64) ALFN1, RN, RD

    
     if (is_saturated) then
            conductivity = K(IM)
     else
           if(X >= 0.0d0) then
              conductivity = K(IM)
            else 
                conductivity = 0.0d0
                N = GAMMAS(IM)
                M = 1.0d0-1.0d0/N

                B = ALFA(IM)*abs(X)
                ALFN = B**N

                SE = (1.0d0+ALFN)**(-M)
                conductivity = K(IM) * SE**(0.5d0) * (1.0d0 - (1.0d0 - SE**(1/M))**M)**2 

                ! ALFN1 = B**(N-1)
              !   RN = (1.0d0 - ALFN1*(1.0d0+ALFN)**(-M))**2
              !   RD = (1.0d0+ALFN)**(M/2)
              ! conductivity = K(IM)*RN/RD
            end if
      end if

    end function FUN_K
    
    function FUN_C(X,ALFA, GAMMAS, TETAS, TETAR, IM) result(capacity)
    real(real64), intent(in) :: X, ALFA(:), GAMMAS(:), TETAS(:), TETAR(:)
    integer, intent(in) :: IM 
    real(real64) capacity
    real(real64) A, N, M, A1, A2, RN, RD 
    
    if (X >= 0.0d0) then
        capacity = 0.0d0
    else if (X <= -1.0d35) then
        capacity = 0.0d0
    else
        capacity = 0.0d0
        N = GAMMAS(IM)
        M = 1.0d0 - 1.0d0/N
        A = ALFA(IM)*abs(X)
        A1 = A**N
        A2 = A**(N - 1.0d0) !=Se
        RN = N*M*(TETAS(IM)-TETAR(IM))*A2*ALFA(IM)
        RD = (1.0d0+A1)**(M+1.0d0)
        capacity = RN/RD
    end if

    end function FUN_C
    
    
    function FUN_T(X, ALFA, GAMMAS, TETAS, TETAR, IM) result(theta)
    real(real64), intent(in) :: X, ALFA(:), GAMMAS(:), TETAS(:), TETAR(:)
    integer, intent(in) :: IM 
    real(real64) theta
    real(real64) N, M, ALFN, RD
    
    if (X >= 0.0d0) then
        theta = TETAS(IM)
    else if(X  <= -1.0d4) then 
        theta = TETAR(IM)
    else      
        theta = 0.0d0
        N = GAMMAS(IM)
        M = 1.0d0 - 1.0d0/N
        ALFN = (ALFA(IM)*abs(X))**N
        RD = (1.0d0+ALFN)**M
        theta = (TETAS(IM)-TETAR(IM))/RD + TETAR(IM)
    end if

    end function FUN_T


end module


PROGRAM main
    use, intrinsic :: iso_fortran_env
    use subprogs

    ! logical ex_A, ex_B
    ! logical is_carbendazim, is_saturated
    ! integer operation_loadingRate, operation_conc, i, j
    real(real64) error_data, KD, Ks
    real(real64) pars(2)
    integer(int32) count_start, count_end, count_rate !, num
    real(real64) calc_time

    ! The following two commands works only for gfortran compiler
    ! inquire(file='./output-pesticide-stageA-10days/.', exist=ex_A)
    ! inquire(file='./output-pesticide-stageB-10days/.', exist=ex_B)


    ! if (ex_A) then
    !     if (ex_B) then
    !         print*, 'Necessary directories already there'
    !         print*, 'Delete csv files inside both directories? [Yes: 1, No: 2]'
    !         read*, num

    !         if (num == 1) then
    !             print*, ''
    !             call system ('cd output-pesticide-stageA-10days && rm -f *.csv')
    !             call system ('cd output-pesticide-stageB-10days && rm -f *.csv')
    !             ! Use the followings in the Windows system instead
    !             ! call system ('cd output-pesticide-stageA && del *.csv')
    !             ! call system ('cd output-pesticide-stageB && del *.csv')

    !         else
    !             call system('cd output-pesticide-stageA-10days && mkdir backup && mv *.csv ./backup')
    !             call system('cd output-pesticide-stageB-10days && mkdir backup && mv *.csv ./backup')
    !         end if


    !     else
    !         print*, 'Creating a necessary directory'
    !         call system('mkdir output-pesticide-stageB-10days' ) 
    !     end if
    ! else       
    !     print*, 'Creating necessary directories'
    !     call system('mkdir output-pesticide-stageA-10days' ) 
    !     call system('mkdir output-pesticide-stageB-10days' ) 

    ! end if

    ! ##### Operational condtions #####
    ! Arguments for solute_flush()  

    ! 1) Target pesticide: carbendazim or chlorothalonil
    ! is_carbendazim = .true. || .false.
    
    ! 2) Is the system saturated or unsaturated?
    ! is_saturated = .true. || .false

    ! 3) Loading rate: 1: 100 L/d, 2: 200 L/d, 3: 400 L/d, 4: 50 L/d
    ! operation_loadingRate = 1 || 2 || 3 || 4  

    ! 4) Influent concentration of target pesticide (chlorothalonil)
    ! 1: 76 ug/L, 2: 79 ug/L, 3: 72 ug/L, 
    ! 4: 236 ug/L, 5: 217 ug/L, 6: 392 ug/L, 7: 0.05 g/L 
    ! operation_conc = 1 || 2 || 3 || 4 || 5 || 6 || 7

    ! 4-2) Actual influent concentration of target pesticide (carbendazim)
    ! 1: 100 ug/L, 2: 73 ug/L, 3: 43 ug/L, 4: 190 ug/L, 5: 127 ug/L
    ! operation_conc = 1 || 2 || 3 || 4 || 5 
    
    ! ##############################################################
    ! Using nested do loops to run simulation for all possible cases
    ! ##############################################################

    error_data = 0.0d0
    open(unit=58, file='./error_data.csv')

    call system_clock(count_start, count_rate)

    ! ### Carbendazim - 7 conditions ###

    ! Carbendazim, saturated, 100 L/d, 10 ug/L
    ! Actual concentraiton 73 ug/L

print*, "Run # 1-1 (after calibration)"
    KD=0.1d0  ! Distibution coefficient
    Ks=7.0d-7 ! First-order degradation rate constant
    pars = [KD, Ks]
    call solute_flush (.true., .true., 1, 2, error_data, pars)
    print*, 'error data after run #1-1', error_data
    write(58, 100) error_data
    error_data = 0.0d0
    
print*, "Run # 1-2 (before calibration)"
    KD=0.58d0  ! Distibution coefficient
    pars = [KD, Ks]
    call solute_flush (.true., .true., 1, 2, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

    ! Carbendazim, saturated, 200 L/d, 10 ug/L
    ! Actual concentration 43 ug/L
print*, "Run # 2-1 (after calibration)"
    Ks = 4.5d-6 
    pars = [KD, Ks]
    call solute_flush (.true., .true., 2, 3, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

print*, "Run # 2-2 (before calibration)"
    Ks = 7.0d-7 
    pars = [KD, Ks]
    call solute_flush (.true., .true., 2, 3, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

    ! Carbendazim, saturated, 400 L/d, 10 ug/L
    ! Actual concentration 190 ug/L
print*, "Run # 3-1 (after calibration)"
    Ks = 9.0d-6 
    pars = [KD, Ks]
    call solute_flush (.true., .true., 3, 4, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

print*, "Run # 3-2 (before calibration)"
    Ks = 7.0d-7 
    pars = [KD, Ks]
    call solute_flush (.true., .true., 3, 4, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

    ! Carbendazim, saturated, 100 L/d, 100 ug/L
    ! Actual concentration 98 ug/L
print*, "Run # 4-1 (after calibration)"
    KD=0.1d0  ! Distibution coefficient
    pars = [KD, Ks]
    call solute_flush (.true., .true., 1, 1, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

print*, "Run # 4-2 (before calibration)"
    KD=0.58d0  ! Distibution coefficient
    pars = [KD, Ks]
    call solute_flush (.true., .true., 1, 1, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

    ! Carbendazim, saturated, 200 L/d, 100 ug/L
    ! Actual concentration 100 ug/L
print*, "Run # 5-1 (after calibration)"
    Ks = 4.5d-6 
    pars = [KD, Ks]
    call solute_flush (.true., .true., 2, 1, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

print*, "Run # 5-2 (before calibration)"
    Ks = 7.0d-7 
    pars = [KD, Ks]
    call solute_flush (.true., .true., 2, 1, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

    ! Carbendazim, saturated, 400 L/d, 100 ug/L
    ! Actual concentration 100 ug/L
print*, "Run # 6-1 (after calibration)"
    Ks = 9.0d-6 
    pars = [KD, Ks]
    call solute_flush (.true., .true., 3, 1, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

print*, "Run # 6-2 (before calibration)"
    Ks = 7.0d-7 
    pars = [KD, Ks]
    call solute_flush (.true., .true., 3, 1, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

    ! Carbendazim, UNsaturated, 200 L/d, 100 ug/L
    ! Actual concentration 127 ug/L

print*, "Run # 7 (no calibration required)"
    pars = [KD, Ks]
    call solute_flush (.true., .false., 2, 5, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0


    ! ### Chlorothalonil - 7 conditions ###

    ! Chlorothalonil, unsaturated, 50 L/d, 100 ug/L
print*, "Run # 8-1 (after calibration)"
    Ks = 1.0d-6
    KD = 2.6d0 
    pars = [KD, Ks]
    call solute_flush (.false., .false., 4, 1, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

print*, "Run # 8-2 (before calibration)"
    KD = 6.5d0 
    pars = [KD, Ks]
    call solute_flush (.false., .false., 4, 1, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

    ! Chlorothalonil, unsaturated, 200 L/d, 100 ug/L
print*, "Run # 9-1 (after calibration)"  
    KD = 4.0d0  
    pars = [KD, Ks]
    call solute_flush (.false., .false., 2, 2, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

print*, "Run # 9-2 (before calibration)"  
    KD = 6.5d0  
    pars = [KD, Ks]
    call solute_flush (.false., .false., 2, 2, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

    ! Chlorothalonil, unsaturated, 400 L/d, 100 ug/L
print*, "Run # 10-1 (after calibration)"
    KD = 6.5d0 
    pars = [KD, Ks]
    call solute_flush (.false., .false., 3, 3, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

print*, "Run # 10-2 (after calibration2)"
    KD = 4.0d0 
    pars = [KD, Ks]
    call solute_flush (.false., .false., 3, 3, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

    ! Chlorothalonil, unsaturated, 50 L/d, 500 ug/L
print*, "Run # 11-1 (after calibration)"
    KD = 2.6d0 
    pars = [KD, Ks]
    call solute_flush (.false., .false., 4, 4, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

print*, "Run # 11-2 (before calibration)"
    KD = 6.5d0 
    pars = [KD, Ks]
    call solute_flush (.false., .false., 4, 4, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

    ! Chlorothalonil, unsaturated, 200 L/d, 500 ug/L
print*, "Run # 12-1 (after calibration)"
    KD = 4.0d0  
    pars = [KD, Ks]
    call solute_flush (.false., .false., 2, 5, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

print*, "Run # 12-2 (before calibration)"
    KD = 6.5d0  
    pars = [KD, Ks]
    call solute_flush (.false., .false., 2, 5, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

    ! Chlorothalonil, unsaturated, 400 L/d, 500 ug/L
print*, "Run # 13-1 (after calibration)"
    KD = 6.5d0 
    pars = [KD, Ks]
    call solute_flush (.false., .false., 3, 6, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

print*, "Run # 13-2 (before calibration2)"
    KD = 4.0d0 
    pars = [KD, Ks]
    call solute_flush (.false., .false., 3, 6, error_data, pars)
    write(58, 100) error_data
    error_data = 0.0d0

    ! Chlorothalonil, unsaturated, 50 L/d, 0.05 g/L
print*, "Run # 14-1 (after calibration)"
    KD = 2.6d0 
    pars = [KD, Ks]
    call solute_flush (.false., .false., 4, 7, error_data, pars)
    write(58, 100) error_data

print*, "Run # 14-2 (before calibration)"
    KD = 6.5d0 
    pars = [KD, Ks]
    call solute_flush (.false., .false., 4, 7, error_data, pars)
    write(58, 100) error_data

    call system_clock(count_end)
    calc_time = dble(count_end - count_start)/dble(count_rate)
    print * , 'Run time = ', calc_time

    close(58)
100 format(F8.3)  
end program